/*
  dethorn.cpp - Translate thorn files to lua.

  Jason Hood, 22 June, 2010.

  5 January, 2011:
  * dump more than one file.

  29 February, 2012:
  * set and use FDUMP.

  5 to 8 March, 2012:
  * set data path based on program location;
  * scratch that, use current directory;
  * format a little nicer (imho).

  v1.01, 4 May, 2023:
  - WriteFile was returning the wrong length, resulting in missing entries.
*/

#define PVERS	"1.01"
#define PDATE	"4 May, 2023"

#include "common.h"
#include "dacom.h"
#include "dalib.h"
#include "from_gui.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>


// readfile.dll contains no write functions, so fudge one.
struct IFileSystem
{
  virtual bool STDCALL open( LPCSTR name );
  virtual UINT STDCALL inc_ref();
  virtual UINT STDCALL dec_ref();
  virtual void STDCALL func_0C() {};
  virtual void STDCALL func_10() {};
  virtual void STDCALL func_14() {};
  virtual UINT STDCALL WriteFile( LPVOID u1, LPVOID data, UINT len, UINT& written, LPVOID u3 );

  FILE* file;
  UINT	ref;
};


bool IFileSystem::open( LPCSTR name )
{
  file = fopen( name, "wb" );
  bool rc = (file != NULL);
  ref = rc;
  return rc;
}


UINT IFileSystem::inc_ref()
{
  return ++ref;
}


UINT IFileSystem::dec_ref()
{
  if (ref != 0)
  {
    if (--ref == 0)
      fclose( file );
  }
  return ref;
}


UINT IFileSystem::WriteFile( LPVOID, LPVOID data, UINT len, UINT& written, LPVOID )
{
  if (ref != 0)
    written = fwrite( data, 1, len, file );
  else
    written = 0;
  return written;
}


void LoadDLL( LPCSTR dll )
{
  HMODULE mod = LoadLibrary( dll );
  if (mod == NULL)
  {
    FDUMP( 0x100001, "%s(%d) : *** ERROR: Could not load dll %s.\n",
	   __FILE__, __LINE__, dll );
    exit( 1 );
  }
}


int Error( int level, LPCSTR fmt, ... )
{
  va_list args;

  va_start( args, fmt );
  vfprintf( stderr, fmt, args );
  va_end( args );

  return 0;
}


void reformat();


int main( int argc, char* argv[] )
{
  if (argc == 1 || strcmp( argv[1], "/?" ) == 0
		|| strcmp( argv[1], "--help") == 0)
  {
    puts( "Dethorn by Jason Hood <jadoxa@yahoo.com.au>.\n"
	  "Version " PVERS " (" PDATE ").  Freeware.\n"
	  "http://freelancer.adoxa.vze.com/\n"
	  "\n"
	  "Translate thorn (*.thn) to lua (*.lua) using Freelancer's own code.\n"
	  "\n"
	  "dethorn FILE...\n"
	  "\n"
	  "FILE is any thorn (or lua) file.  *.thn becomes *.lua; anything else will\n"
	  "append .lua (e.g. file.thn -> file.lua; file.ext -> file.ext.lua)."
	);
    from_gui_tip = true;
    return 0;
  }

  FDUMP = Error;

  LoadDLL( "thorn.dll" );
  LoadDLL( "readfile.dll" );
  LoadDLL( "x86math.dll" );

  DALib::Dacom = DACOM_Acquire();

  /*
  char* datapath = strdup( _pgmptr );
  char* slash = strrchr( datapath, '\\' );
  strcpy( slash - 3, "DATA" );
  DWORD attrib = GetFileAttributes( datapath );
  if (attrib == INVALID_FILE_ATTRIBUTES || !(attrib & FILE_ATTRIBUTE_DIRECTORY))
  {
    FDUMP( 0x100001, "%s(%d): *** ERROR: Not a directory %s.\n",
	   __FILE__, __LINE__, datapath );
    return 1;
  }
  DALib::OpenData( datapath );
  */
  DALib::OpenData( "." );

  int rc = 0;
  IFileSystem* fs = new IFileSystem;
  char lua[MAX_PATH];

  reformat();

  FILE* con = NULL;
  int	con_len = 0;
  if (--argc > 1)
  {
    con = fopen( "con", "w" );
    setbuf( con, NULL );
  }
  for (int i = 1; *++argv; ++i)
  {
    if (con)
      con_len = fprintf( con, "%d/%d\r", i, argc );
    IScriptEngine* script = ThornScriptLoad( *argv );
    if (script == NULL)
    {
      rc = 2;
      continue;
    }
    size_t len = strlen( *argv );
    if (len > 4 && _stricmp( *argv + len - 4, ".thn" ) == 0)
      _snprintf( lua, MAX_PATH, "%.*slua", len - 3, *argv );
    else
      _snprintf( lua, MAX_PATH, "%s.lua", *argv );
    bool open = fs->open( lua );
    /*
    if (!open && GetLastError() == ERROR_PATH_NOT_FOUND)
    {
      _snprintf( lua, MAX_PATH, "%s\\%s.lua", datapath, *argv );
      open = fs->open( lua );
    }
    */
    if (open)
    {
      rc = script->dump( fs );
      fs->dec_ref();
    }
    else
    {
      FDUMP( 0x100001, "%s(%d) : *** ERROR: Could not create %s.\n",
	     __FILE__, __LINE__, lua );
      rc = 3;
    }

    ThornScriptDestroy( &script );
  }
  if (con_len)
  {
    fprintf( con, "%*c", con_len, '\r' );
    fclose( con );
  }

  DALib::CloseData();

  return rc;
}


__declspec(naked)
void identity()
{
  static const float im[3][3] = { { 1, 0, 0 }, { 0, 1, 0 }, { 0, 0, 1 } };
  static const char is[] = "\t\t\torient = { { 1, 0, 0 }, { 0, 1, 0 }, { 0, 0, 1 } }\r\n";

  __asm {
	push	ecx
	push	edi
	push	esi
	mov	esi, [esp+8+12]
	mov	edi, offset im
	mov	ecx, 9
	repe	cmpsd
	jne	done
	mov	dword ptr [esp+4+12], offset is
  done:
	pop	esi
	pop	edi
	pop	ecx
	push	0x6F2CF60
	ret
  }
}


__declspec(naked)
void lightprops()
{
  static const char brace[] = "\t\t\t}";

  __asm {
	lea	ecx, [esp+0x14]
	push	offset brace
	push	0x6F2BEC2
	ret
  }
}


void reformat()
{
  DWORD dummy;
  VirtualProtect( (LPVOID)0x6F2A000, 0x4000, PAGE_EXECUTE_READWRITE, &dummy );
  VirtualProtect( (LPVOID)0x6F5E000, 0x2000, PAGE_READWRITE, &dummy );

  // Use %.7g instead of %.6f.
  *(short*)0X6F5F4C1 =		// pos
  *(short*)0X6F5F4C7 =
  *(short*)0X6F5F4CD =
  *(short*)0x6F5F43E =		// sparam
  *(short*)0x6F5F40C =		// floor_height
  *(short*)0x6F5F3D4 =		// fovh
  *(short*)0x6F5F3C0 =		// hvaspect
  *(short*)0x6F5F3A9 =		// nearplane
  *(short*)0x6F5F390 =		// farplane
  *(short*)0x6F5F35E =		// various audioprops
  *(short*)0x6F5F6B4 =		// various lightprops
  *(short*)0x6F5F6BA =
  *(short*)0x6F5F6C0 =
  *(short*)0x6F5F676 =		// various
  *(short*)0x6F5EE76 =		// various event values
  *(short*)0x6F5EE48 =		// spatialprops.pos
  *(short*)0x6F5EE4E =
  *(short*)0x6F5EE54 =
  *(short*)0x6F5EE2B =		// spatialprops.q_orient
  *(short*)0x6F5EE31 =
  *(short*)0x6F5EE37 =
  *(short*)0x6F5EE3D =
		       0x6737;	// "7g"

  // Change orient to multiple lines, with an extra space for the plus sign.
  // This assumes a tab size of 4.
  *(char**)0x6F2AD80 = "\t\t\torient = { { % f, % f, % f },\r\n"
			   "\t\t\t\t\t   { % f, % f, % f },\r\n"
			   "\t\t\t\t\t   { % f, % f, % f } }\r\n";

  // Add some spaces around the ambient and color values.
  *(char**)0x6F2AF93 = ",\r\n\t\tambient = { %d, %d, %d }";
  memcpy( (LPVOID)0x6F5F67C, "{ %d, %d, %d },", 15 );

  // Remove now-unused trailing space from lightprops values.
  *(char*)0x6F5F679 =
  *(char*)0x6F5F68F =
  *(char*)0x6F5F6C5 = '\0';

  // Adjust atten.
  memcpy( (LPVOID)0x6F5F694, "{ %.7g, %.7g, %.7g }", 21 );

  // Patch in code to handle the identity matrix separately.
  *(DWORD*)0x6F2AD89 = (DWORD)identity - 0x6F2AD89 - 4;

  // Remove the extra line on pathprops.
  *(char*)0x6F5F2E1 = '\0';

  // Do events on multiple lines.
  *(char**)0x6F2B247 = "{\r\n\t\t%.3f";
  *(char**)0x6F2B273 = ", { \"%s\"";
  *(char**)0x6F2B28D = ", \"%s\" }";
  *(char**)0x6F2B299 = " }";
  *(short*)0x6F5F23D = 0x0A0D;		// add a blank line between events

  // This starts most of the events.
  *(char**)0x6F2B3E6 =
  *(char**)0x6F2B53F =
  *(char**)0x6F2B579 =
  *(char**)0x6F2B68C =
  *(char**)0x6F2B6E8 =
  *(char**)0x6F2B858 =
  *(char**)0x6F2B994 =
  *(char**)0x6F2B9E1 =
  *(char**)0x6F2BB27 =
  *(char**)0x6F2BCAC =
  *(char**)0x6F2BF0B =
  *(char**)0x6F2BF7C =
  *(char**)0x6F2BFB6 =
  *(char**)0x6F2C02B =
  *(char**)0x6F2C0AB =
  *(char**)0x6F2C12B = ",\r\n\t\t{\r\n\t\t\tduration = %.3f";

  // START_MOTION
  *(char**)0x6F2B2BF = ",\r\n\t\t{\r\n\t\t\tanimation = \"%s\"";
  *(char**)0x6F2B2D5 = ",\r\n\t\t\tlocked_bone = \"%s\"";
  *(char**)0x6F2B2F4 = ",\r\n\t\t\tduration = %.3f";
  *(char**)0x6F2B312 = ",\r\n\t\t\tstart_time = %.3f";
  *(char**)0x6F2B334 = ",\r\n\t\t\ttrans_time = %.3f";
  *(char**)0x6F2B356 = ",\r\n\t\t\ttime_scale = %.7g";
  *(char**)0x6F2B378 = ",\r\n\t\t\tweight = %.7g";
  *(char**)0x6F2B38F = ",\r\n\t\t\theading = %.7g";
  *(char**)0x6F2B3AA = ",\r\n\t\t\ttrans_scale = %.7g";
  *(char**)0x6F2B3C2 =
  *(char**)0x6F2B668 =
  *(char**)0x6F2B6C4 =
  *(char**)0x6F2BB03 =
  *(char**)0x6F2BECE =
  *(char**)0x6F2C18C = ",\r\n\t\t\tflags = USE_SCRIPT_DURATION";

  // START_SOUND
  *(char**)0x6F2B3F8 =
  *(char**)0x6F2BF29 = ",\r\n\t\t\tstart_time = %d";
  *(char**)0x6F2B410 =
  *(char**)0x6F2B77B =
  *(char**)0x6F2B8D4 =
  *(char**)0x6F2BF3D = ",\r\n\t\t\tflags =";

  // USER_EVENT (not used, won't bother)

  // SUBTITLE (not used, but perhaps it could be)
  *(char**)0x6F2B555 = ",\r\n\t\t\tstrid = %u";

  // START_IK
  *(char**)0x6F2B587 = ",\r\n\t\t\tend_effector = \"%s\"";
  *(char**)0x6F2B599 = ",\r\n\t\t\tcount_to_root = %d";
  *(char**)0x6F2B5AF =
  *(char**)0x6F2B73E =
  *(char**)0x6F2BB39 =
  *(char**)0x6F2C13D = ",\r\n\t\t\ttarget_part = \"%s\"";
  *(char**)0x6F2B5CA =
  *(char**)0x6F2B759 =
  *(char**)0x6F2BB54 =
  *(char**)0x6F2C158 = ",\r\n\t\t\ttarget_type = ROOT";
  *(char**)0x6F2B5D1 =
  *(char**)0x6F2B760 =
  *(char**)0x6F2BB5B =
  *(char**)0x6F2C15F = ",\r\n\t\t\ttarget_type = PART";
  *(char**)0x6F2B5D8 =
  *(char**)0x6F2B767 =
  *(char**)0x6F2BB62 =
  *(char**)0x6F2C166 = ",\r\n\t\t\ttarget_type = HARDPOINT";
  *(char**)0x6F2B5E6 = ",\r\n\t\t\tdamping = %.7g";
  *(char**)0x6F2B5FF =
  *(char**)0x6F2B70F =
  *(char**)0x6F2B8A3 = ",\r\n\t\t\tup = %s";
  *(char**)0x6F2B618 =
  *(char**)0x6F2B728 =
  *(char**)0x6F2B8BC = ",\r\n\t\t\tfront = %s";
  *(char**)0x6F2B62B = ",\r\n\t\t\tpoint_at = %d";
  *(char**)0x6F2B63E = ",\r\n\t\t\tmove_to = %d";
  *(char**)0x6F2B650 = ",\r\n\t\t\ttransition_duration = %.3f";

  // CONNECT_HARDPOINT
  *(char**)0x6F2B69A = ",\r\n\t\t\thardpoint = \"%s\"";
  *(char**)0x6F2B6AC = ",\r\n\t\t\tparent_hardpoint = \"%s\"";

  // ATTACH_ENTITY
  *(char**)0x6F2B6F6 =
  *(char**)0x6F2B88A = ",\r\n\t\t\toffset = { %.7g, %.7g, %.7g }";

  // START_PATH_ANIMATION
  *(char**)0x6F2B866 = ",\r\n\t\t\tstart_percent = %.7g";
  *(char**)0x6F2B878 = ",\r\n\t\t\tstop_percent = %.7g";
  *(char**)0x6F2B970 =
  *(char**)0x6F2BF92 =
  *(char**)0x6F2BFC8 =
  *(char**)0x6F2C03D =
  *(char**)0x6F2C0BD =
  *(char**)0x6F2C1A2 = ",\r\n\t\t\tevent_flags = %d";

  // START_LIGHT_PROP_ANIM
  *(char**)0x6F2B99E = ",\r\n\t\t\tlightprops =\r\n\t\t\t{\r\n";
  *(PDWORD)0x6F2B9B9 = 0x6F5F2E8;
  *(char**)0x6F2B9BE = "\t\t\t\t";
  // Patch in code to remove an extra newline.
  *(DWORD*)0x6F2B9CB = (DWORD)lightprops - 0x6F2B9CB - 4;

  // START_CAMERA_PROP_ANIM
  *(char**)0x6F2B9EB = ",\r\n\t\t\tcameraprops =\r\n\t\t\t{\r\n\t\t\t\t";
  *(char**)0x6F2BA35 =
  *(char**)0x6F2BA76 =
  *(char**)0x6F2BAB7 =
  *(char**)0x6F2BBCD =
  *(char**)0x6F2BC3A =
  *(char**)0x6F2BD01 =
  *(char**)0x6F2BD41 =
  *(char**)0x6F2BD82 =
  *(char**)0x6F2BDC3 =
  *(char**)0x6F2BE03 =
  *(char**)0x6F2BE44 =
  *(char**)0x6F2BE86 = ",\r\n\t\t\t\t";

  // START_SPATIAL_PROP_ANIM
  *(char**)0x6F2BB6C = ",\r\n\t\t\tspatialprops =\r\n\t\t\t{\r\n\t\t\t\t";
  // Use original q_orient values (not that it seems to matter).
  memcpy( (LPVOID)0x6F2BBF3, "\x8D\x55\x60"             // lea  edx, [ebp+0x60]
			     "\x52"                     // push edx
			     "\x68\x1C\xEE\xF5\x06"     // push 0x6F5EE1C
			     "\x8D\x4C\x24\x1C"         // lea  ecx, [esp+0x1C]
			     "\xE8\x1B\x14\x00\x00"     // call 0x6F2D020
			     "\xEB\x28"                 // jmp  0x6F2BC2F
			     , 20 );
  *(char**)0x6F2BC22 = "orient = { { % f, % f, % f },\r\n"
		   "\t\t\t\t\t\t   { % f, % f, % f },\r\n"
		   "\t\t\t\t\t\t   { % f, % f, % f } }";
  *(char**)0x6F2BC61 = "{ %.7g, ";
  *(char**)0x6F2BC7A = "%s }";
  *(char**)0x6F2BAEB =
  *(char**)0x6F2BC88 =
  *(char**)0x6F2BEBE =
  *(char**)0x6F2C0EE = "\r\n\t\t\t}";

  // START_AUDIO_PROP_ANIM
  *(char**)0x6F2BCB6 = ",\r\n\t\t\taudioprops =\r\n\t\t\t{\r\n\t\t\t\t";

  // START_PSYS
  *(char**)0x6F2BFD6 = ",\r\n\t\t\tpsysprops =\r\n\t\t\t{\r\n\t\t\t\t";

  // START_FOG_PROP_ANIM
  *(char**)0x6F2C04B = ",\r\n\t\t\tfogprops =\r\n\t\t\t{\r\n\t\t\t";
  *(char**)0x6F2C3FB =
  *(char**)0x6F2C42F =
  *(char**)0x6F2C465 =
  *(char**)0x6F2C48D =
  *(char**)0x6F2C4BD =
  *(char**)0x6F2C4E5 =
  *(char**)0x6F2C50D = "\t%s = ";
  *(char**)0x6F2C417 =
  *(char**)0x6F2C44D =
  *(char**)0x6F2C4A6 = "%s,\r\n\t\t\t";
  *(char**)0x6F2C475 = "{ %d, %d, %d },\r\n\t\t\t";
  *(char**)0x6F2C4CD =
  *(char**)0x6F2C4F5 = "%.7g,\r\n\t\t\t";
  *(char**)0x6F2C51D = "%.7g\r\n\t\t\t";

  // START_REVERB_PROP_ANIM (not used, incomplete)
  *(char**)0x6F2C0CB = ",\r\n\t\t\treverbprops =\r\n\t\t\t{\r\n\t\t\t\t";

  // START_FLR_HEIGHT_ANIM
  *(char**)0x6F2C174 = ",\r\n\t\t\tfloor_height = %.7g";

  // common
  *(char**)0x6F2C1C5 = ",\r\n\t\t\tpcurve_period = %d";
  *(char**)0x6F2D161 = ",\r\n\t\t\tparam_curve =\r\n\t\t\t{\r\n"
		       "\t\t\t\tCLSID = \"%s\",\r\n"
		       "\t\t\t\tpoints =\r\n\t\t\t\t{\r\n\t\t\t\t";
  *(char**)0x6F2D1E1 = "\t{ %9f, %9f, %9f, %9f },\r\n\t\t\t\t";
  *(char**)0x6F2D25A = "}\r\n\t\t\t}";
  *(char**)0x6F2C1D3 = "\r\n\t\t}";
  *(char**)0x6F2C1DD = "\r\n\t}";
}
